/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MPIA, NEVEC - MIDI project
*
* "@(#) $Id: oi_vis.c,v 1.9 2010/01/31 11:51:37 jaffe Exp $"
*
* who        when      what
* ---------  --------  ----------------------------------------------
* jdejong    2002-05-08 imported
*
*/

#define _POSIX_SOURCE_1
#include "vltPort.h"

/* $Id: oi_vis.c,v 1.9 2010/01/31 11:51:37 jaffe Exp $ */

/*----------------------------------------------------------------------- */
/* Visibility data FITS binary tables routines                            */
/* Partially/fully calibrated data table for optical/IR interferometer    */
/*----------------------------------------------------------------------- */
#include "oi_vis.h"

/* internal prototypes */
/* get data from table_def specific to visibility table */
void get_vis_info (vis_table_def *table);
/* file global variables for table type */
static char *tabletype = "OI_VIS";


/* vis_row functions */
vis_table_row* make_vis_row(vis_table_def* table)
/*----------------------------------------------------------------------- */
/*   Make an vis_table_row structure                                   */
/*   Inputs:                                                              */
/*      table     pointer to vis table structure                 */
/*   Returns:                                                             */
/*      An vis_table_row structure pointer ; NULL indicates failure    */
/*      Data pointers set to generic structure values                     */
/*----------------------------------------------------------------------- */
{
  vis_table_row *out;
  int i, icol, nregion;

  /* create basic structure */
  out = (vis_table_row*)malloc(sizeof(vis_table_row));
  if (out==NULL) return out; /* allocation failure? */
  
  /* zero memory */
  memset (out, 0, sizeof(vis_table_row));
 
  /* allocate internal arrays */
  out->visAmp = (double *)malloc(table->nWave * sizeof(double));
  memset (out->visAmp, 0, table->nWave*sizeof(double));
  out->visAmpErr = (double *)malloc(table->nWave * sizeof(double));
  memset (out->visAmpErr, 0, table->nWave*sizeof(double));
  out->visPhi = (double *)malloc(table->nWave * sizeof(double));
  memset (out->visPhi, 0, table->nWave*sizeof(double));
  out->visPhiErr = (double *)malloc(table->nWave * sizeof(double));
  memset (out->visPhiErr, 0, table->nWave*sizeof(double));
  out->flag = (char *)malloc(table->nWave * sizeof(char));
  memset (out->flag, 0, table->nWave*sizeof(char));

  /* Make generic structure */
  out->raw = make_table_row_from_def(table->def);

  return out; /* done */
} /* end  make_vis_table_row */

void kill_vis_row (vis_table_row *row)
/*----------------------------------------------------------------------- */
/*   Delete an vis_table_row structure                                   */
/*   Inputs:                                                              */
/*      row    pointer to the vis_table_row structure.                   */
/*----------------------------------------------------------------------- */
{
  int i, nregion;
  if (row==NULL) return; /* sanity check */

  free (row->raw); /* generic structure */
  free (row);      /* basic structure */
} /* end  kill_vis_row */


/* vis_table functions */
vis_table_def*
make_vis_table_def(fitsfile *fptr, int nWave, char *date_obs, char* arrname,
   char *insname, int *ierr)
/*----------------------------------------------------------------------- */
/*   Make a new vis_table in an open FITS file                            */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      nWave     number of entries per visibility point                  */
/*      date_obs  date of observation                                     */
/*      arrname   array name                                              */
/*      insname   instrument/disperser name                               */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An vis_table_def pointer; NULL indicates failure                  */
/*----------------------------------------------------------------------- */
{
  vis_table_def *out = NULL;
  int    version = 1; /* table version number */
  int    ikey, icol, i;
  long   nx, ny;

  if (*ierr>0) return NULL;    /* existing error condition? */

  if (fptr==NULL) return NULL; /* sanity check */

  /* create output object */
  out = (vis_table_def*)malloc(sizeof(vis_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(vis_table_def)); /* zero fill */

  /* create table_def */
  out->def = make_table_def(OI_VIS_TABLE, 0, 0);
  if (out->def==NULL) { /* check for failure */
    kill_vis_table_def(out); /* clean up */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating OI_VIS_TABLE for %s table_def \n", 
	    tabletype);
    return NULL;
  }

  /* add keywords */
  ikey = table_find_key(out->def, "OI_REVN");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)&version, ierr);
  ikey = table_find_key(out->def, "DATE-OBS");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)date_obs, ierr);
  ikey = table_find_key(out->def, "ARRNAME");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)arrname, ierr);
  ikey = table_find_key(out->def, "INSNAME");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)insname, ierr);

  out->nWave = nWave;
  /* get info from table descriptor */
  get_vis_info (out);

  /* set data size */
  icol = out->visamp_col; 
  out->def->colcount[icol] =  nWave;
  icol = out->visamperr_col; 
  out->def->colcount[icol] =  nWave;
  icol = out->visphi_col; 
  out->def->colcount[icol] =  nWave;
  icol = out->visphierr_col; 
  out->def->colcount[icol] =  nWave;
  icol = out->flag_col; 
  out->def->colcount[icol] =  nWave;

  /* open imaging data table */
  make_table_from_def (fptr, out->def, ierr);

  return out; /* done */
} /* end make_vis_table */

vis_table_def* open_vis_table(fitsfile *fptr, int *ierr)
/*----------------------------------------------------------------------- */
/*   Open an existing vis_table in the active FITS file            */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An vis_table_def pointer; NULL indicates failure           */
/*----------------------------------------------------------------------- */
{
  vis_table_def *out = NULL;
  int i,iKey;
  char kString[20], typeString[20];
  char *tForm = "TFORM";
  int nWave, found;

  if (*ierr>0) return NULL;    /* existing error condition? */
  if (fptr==NULL) return NULL; /* sanity check */

  /* create output object */
  out = (vis_table_def*)malloc(sizeof(vis_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(vis_table_def)); /* zero fill */

  /* position FITS file at  table */
  if (fits_movnam_hdu(fptr, BINARY_TBL, tabletype, 0, ierr)) {
    fprintf(stderr,"ERROR %d finding %s table \n", *ierr, tabletype);
    kill_vis_table_def(out); /* clean up */
    *ierr = 10000;
    return NULL;
  }

  /* get nWave by reading format data in header ugh! */
  found = 0;
  for (i=1; i<20 && (found==0); i++) {
     sprintf(kString, "TTYPE%d",i);
     iKey = fits_read_key(fptr, TSTRING, kString, (void*)typeString, 0, ierr);
     if (strncmp(typeString, "VISAMP", 6)==0) {
        sprintf(kString, "TFORM%d",i);
        iKey = fits_read_key(fptr, TSTRING, kString, (void*)typeString, 0,  ierr);
        sscanf(typeString, "%d", &nWave);
        found = 1;
     }
  }
  out->nWave = nWave;
  /* get table descriptor */
  out->def = make_table_def_from_table (fptr);
  if (out->def==NULL) { /* error check */
    fprintf(stderr,"ERROR making table def for %s \n", tabletype);
    kill_vis_table_def(out); /* clean up */
    *ierr = 10000;
    return NULL;
  }

  /* copy info from table descriptor */
  get_vis_info (out);

  return out; /* done */
} /* end open_vis_table */

int read_vis_table_row(fitsfile *fptr, vis_table_def* table, 
			    vis_table_row *row, long row_number,
			    int *ierr)
/*----------------------------------------------------------------------- */
/*   Read the next row from an vis_table                           */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row_number which row (1-rel) in the table                         */   
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      row       pointer to an extant row structure                      */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  table_row *trow;
  int        i;

  if (*ierr>0) return 1;     /* existing error condition? */
  if (fptr==NULL) return 1;  /* sanity check */
  if (table==NULL) return 1; /* sanity check */
  if (row==NULL) return 1;   /* sanity check */

  /* set row number */
  if (row_number>0) { /* explicit row number */
    row->raw->trow = row_number;
    table->row = row_number;
  } else { /* next row */
    row->raw->trow = ++table->row;
  }

  /* read row */
  table_read_row(fptr, table->def, row->raw, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR  read_vis_table: reading table %s \n", 
	    tabletype);
    return 1;
  }

  /* copy relevant data from generic table_row */
  trow = row->raw;
  row->targetId    = trow->data[table->target_col]->tdata.tshort[0];
  row->time        = trow->data[table->time_col]->tdata.tdouble[0];
  row->mjd         = trow->data[table->mjd_col]->tdata.tdouble[0];
  row->intTime     = trow->data[table->intTime_col]->tdata.tdouble[0];
  row->ucoord      = trow->data[table->u_col]->tdata.tdouble[0];
  row->vcoord      = trow->data[table->v_col]->tdata.tdouble[0];
  row->station[0]      = trow->data[table->station_col]->tdata.tshort[0];
  row->station[1]      = trow->data[table->station_col]->tdata.tshort[1];

  for (i=0; i<table->nWave; i++) {
     row->visAmp[i] = trow->data[table->visamp_col]->tdata.tdouble[i];
     row->visAmpErr[i] = trow->data[table->visamperr_col]->tdata.tdouble[i];
     row->visPhi[i] = trow->data[table->visphi_col]->tdata.tdouble[i];
     row->visPhiErr[i] = trow->data[table->visphierr_col]->tdata.tdouble[i];
     row->flag[i] = trow->data[table->flag_col]->tdata.tchar[i];
  }
  return 0; /* OK - done */
} /* end read_vis_table */

int write_vis_table_row(fitsfile *fptr, vis_table_def* table, 
			     vis_table_row *row, long row_number, 
			     int *ierr)
/*----------------------------------------------------------------------- */
/*   Write  the next row to an vis_table                         */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row       pointer to an extant row structure                      */
/*      row_number which row (1-rel) in the table                         */     
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  table_row *trow;
  int        i;

  if (*ierr>0) return 1;     /* existing error condition? */
  if (fptr==NULL) return 1;  /* sanity check */
  if (table==NULL) return 1; /* sanity check */
  if (row==NULL) return 1;   /* sanity check */

  /* set row number */
  if (row_number>0) { /* explicit row number */
    row->raw->trow = row_number;
    table->row = row_number;
  } else { /* next row */
    row->raw->trow = ++table->row;
  }

  /* copy relevant data to generic table_row */
  trow = row->raw;
  trow->data[table->target_col]->tdata.tshort[0] = row->targetId;
  trow->data[table->time_col]->tdata.tdouble[0]  =  row->time;
  trow->data[table->mjd_col]->tdata.tdouble[0]   =  row->mjd;
  trow->data[table->intTime_col]->tdata.tdouble[0] = row->intTime;
  trow->data[table->u_col]->tdata.tdouble[0] = row->ucoord;
  trow->data[table->v_col]->tdata.tdouble[0] = row->vcoord;
  trow->data[table->station_col]->tdata.tshort[0] = row->station[0];
  trow->data[table->station_col]->tdata.tshort[1] = row->station[1];

  for (i=0; i<table->nWave; i++) {
     trow->data[table->visamp_col]->tdata.tdouble[i]    = row->visAmp[i];
     trow->data[table->visamperr_col]->tdata.tdouble[i] = row->visAmpErr[i];
     trow->data[table->visphi_col]->tdata.tdouble[i]    = row->visPhi[i];
     trow->data[table->visphierr_col]->tdata.tdouble[i] = row->visPhiErr[i];
     trow->data[table->flag_col]->tdata.tchar[i]       = row->flag[i];
  }
  /* write row */
  table_write_row(fptr, table->def, row->raw, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR  write_vis_table: writing table %s \n", 
	    tabletype);
    return 1;
  }

  return 0; /* OK - done */
} /* end write_vis_table */

int close_vis_table(fitsfile *fptr, vis_table_def* table, 
			     int *ierr)
/*----------------------------------------------------------------------- */
/*   Close an open vis_table, associated structures are destroyed  */
/*   Note: this closes the full FITS file                                 */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  if (*ierr>0) return 1;    /* existing error condition? */
  if (fptr==NULL) return 1; /* sanity check */

  /* do actual file close */
  fits_close_file (fptr, ierr);

  /* delete structures */
  kill_vis_table_def(table);

  return 0; /* OK - done */
} /* end close_vis_table */

void kill_vis_table_def (vis_table_def* table)
/*----------------------------------------------------------------------- */
/*   Delete an vis_table structure                                 */
/*   Inputs:                                                              */
/*      table     table definitions structure from open or create         */
/*----------------------------------------------------------------------- */
{
  if (table==NULL) return; /* sanity check */

  /* table definition structure */
  kill_table_def(table->def);

  /* free basic structure */
  free (table);
} /* end kill_vis_table_def */

/*************************private functions ******************************/

void get_vis_info (vis_table_def *table)
/*----------------------------------------------------------------------- */
/*   Get data from table_def specific to vis table                 */
/*   Inputs:                                                              */
/*      table     table definitions structure                             */
/*----------------------------------------------------------------------- */
{
  if (table==NULL) return;         /* sanity check */

  /* lookup column numbers (0-rel) */
  table->target_col    = table_find_col (table->def, "TARGET_ID")-1;
  table->time_col      = table_find_col (table->def, "TIME")-1;
  table->mjd_col       = table_find_col (table->def, "MJD")-1;
  table->intTime_col   = table_find_col (table->def, "INT_TIME")-1;
  table->visamp_col    = table_find_col (table->def, "VISAMP")-1;
  table->visamperr_col = table_find_col (table->def, "VISAMPERR")-1;
  table->visphi_col    = table_find_col (table->def, "VISPHI")-1;
  table->visphierr_col = table_find_col (table->def, "VISPHIERR")-1;
  table->u_col         = table_find_col (table->def, "UCOORD")-1;
  table->v_col         = table_find_col (table->def, "VCOORD")-1;
  table->station_col   = table_find_col (table->def, "STA_INDEX")-1;
  table->flag_col      = table_find_col (table->def, "FLAG")-1;
} /* end  get_vis_info */

/* */
